package mv3302;

import static java.lang.Double.NaN;
import java.util.SortedSet;
import java.util.TreeSet;
import simkit.Entity;
import static simkit.Priority.HIGH;
import simkit.SimEntityBase;
import simkit.random.RandomVariate;

/**
 * Implementation of the multiple server queue model using<code>Entity</code>
 * objects to represent individual customers. This allows the model to
 * explicitly tally measures such as delay in queue and time in system.
 *
 * @author ahbuss
 */
public class EntityServer extends SimEntityBase {

    /**
     * Number of servers (k)
     */
    private int totalNumberServers;

    /**
     * Generates service times ({t<sub>S</sub>})
     */
    private RandomVariate serviceTimeGenerator;

    /**
     * # of available servers (S)
     */
    protected int numberAvailableServers;

    /**
     * Container of waiting customers (q)
     */
    protected SortedSet<Entity> queue;

    /**
     * time a given customer spends in queue (D - a transient state)
     */
    protected double delayInQueue;

    /**
     * time a given customer spends in the system (W - a transient state)
     */
    protected double timeInSystem;

    /**
     * Total number who complete service
     */
    protected int numberServed;

    /**
     * Instantiate an EntityServer with the given titalNumberServers and
     * serviceTimeGenerator. Note: should call <code>this()</code> to ensure
     * that <code>queue</code> is instantiated
     *
     * @param totalNumberServers
     * @param serviceTimeGenerator
     */
    public EntityServer(int totalNumberServers, RandomVariate serviceTimeGenerator) {
        this();
        setTotalNumberServers(totalNumberServers);
        setServiceTimeGenerator(serviceTimeGenerator);
    }

    /**
     * Instantiate the <code>queue</code>state variable
     */
    public EntityServer() {
        this.queue = new TreeSet<>();
    }

    /**
     * Initialize state variables:<ul>
     * <li>Empty <code>queue</code>
     * <li>set numberAvailableServers to totalNumberServers
     * <li>Set numberServed to 0
     * <li>Set delayInQueue to NaN
     * <li>Set timeInSystem to NaN </ul>
     */
    @Override
    public void reset() {
        super.reset();
        queue.clear();
        numberAvailableServers = getTotalNumberServers();
        numberServed = 0;
        delayInQueue = NaN;
        timeInSystem = NaN;
    }

    /**
     * Fire property changes for time-varying states
     */
    public void doRun() {
        firePropertyChange("queue", getQueue());
        firePropertyChange("numberAvailableServers", getNumberAvailableServers());
        firePropertyChange("numberServed", getNumberServed());
    }

    /**
     * Stamp time and add customer to <code>queue</code><br>
     * if available server, schedule StartService with delay of 0.0
     *
     * @param customer Arriving customer
     */
    public void doArrival(Entity customer) {
        customer.stampTime();
        var oldQueue = getQueue();
        queue.add(customer);
        firePropertyChange("queue", oldQueue, getQueue());

        if (numberAvailableServers > 0) {
            waitDelay("StartService", 0.0, HIGH);
        }
    }

    /**
     * Decrement numbervailableServers<br>
     * Remove first customer from <code>queue</code><br>
     * Compute delayInQueue as elapsed time<br>
     * Schedule EndSrevice with delay generated by serviceTimeGenerator
     */
    public void doStartService() {
        var oldNumberAvailableServers = getNumberAvailableServers();
        numberAvailableServers -= 1;
        firePropertyChange("numberAvailableServers", oldNumberAvailableServers, getNumberAvailableServers());

        var oldQueue = getQueue();
        var customer = queue.first();
        queue.remove(customer);
        firePropertyChange("queue", oldQueue, getQueue());

        delayInQueue = customer.getElapsedTime();
        firePropertyChange("delayInQueue", getDelayInQueue());

        waitDelay("EndService", serviceTimeGenerator, customer);
    }

    /**
     * Increment numberAvailableServers<br>
     * Compute timeInSystem as customer's elapsedTime<br>
     * Increment numberServed<br>
     * If customer(s) in <code>queue</code>, schedule StartService with delay of 0.0
     * 
     * @param customer Given customer completing service 
     */
    public void doEndService(Entity customer) {
        var oldNumberAvailableServers = getNumberAvailableServers();
        numberAvailableServers += 1;
        firePropertyChange("numberAvailableServers", oldNumberAvailableServers, getNumberAvailableServers());

        timeInSystem = customer.getElapsedTime();
        firePropertyChange("timeInSystem", getTimeInSystem());

        var oldNumberServed = getNumberServed();
        numberServed += 1;
        firePropertyChange("numberServed", oldNumberServed, getNumberServed());

        if (!queue.isEmpty()) {
            waitDelay("StartService", 0.0, HIGH);
        }

    }

    /**
     * @return the totalNumberServers
     */
    public int getTotalNumberServers() {
        return totalNumberServers;
    }

    /**
     * @param totalNumberServers the totalNumberServers to set
     * @throws IllegalArgumentException if totalNumberServers &le; 0
     */
    public void setTotalNumberServers(int totalNumberServers) {
        if (totalNumberServers <= 0) {
            throw new IllegalArgumentException("totalNumberServers must be > 0: " + totalNumberServers);
        }
        this.totalNumberServers = totalNumberServers;
    }

    /**
     * @return the serviceTimeGenerator
     */
    public RandomVariate getServiceTimeGenerator() {
        return serviceTimeGenerator;
    }

    /**
     * @param serviceTimeGenerator the serviceTimeGenerator to set
     */
    public void setServiceTimeGenerator(RandomVariate serviceTimeGenerator) {
        this.serviceTimeGenerator = serviceTimeGenerator;
    }

    /**
     * @return the numberAvailableServers
     */
    public int getNumberAvailableServers() {
        return numberAvailableServers;
    }

    /**
     * @return the queue
     */
    public SortedSet<Entity> getQueue() {
        return new TreeSet<>(queue);
    }

    /**
     * @return the numberServed
     */
    public int getNumberServed() {
        return numberServed;
    }

    /**
     * @return the delayInQueue
     */
    public double getDelayInQueue() {
        return delayInQueue;
    }

    /**
     * @return the timeInSystem
     */
    public double getTimeInSystem() {
        return timeInSystem;
    }

}
